from kirin.interp import Successor, MethodTable, AbstractFrame, impl
from kirin.dialects.cf.stmts import Branch, ConditionalBranch
from kirin.analysis.typeinfer import TypeInference
from kirin.dialects.cf.dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(MethodTable):

    @impl(Branch)
    def branch(self, interp: TypeInference, frame: AbstractFrame, stmt: Branch):
        frame.worklist.append(
            Successor(stmt.successor, *frame.get_values(stmt.arguments))
        )
        return ()

    @impl(ConditionalBranch)
    def conditional_branch(
        self, interp: TypeInference, frame: AbstractFrame, stmt: ConditionalBranch
    ):
        frame.worklist.append(
            Successor(stmt.else_successor, *frame.get_values(stmt.else_arguments))
        )
        frame.worklist.append(
            Successor(stmt.then_successor, *frame.get_values(stmt.then_arguments))
        )
        return ()



import io

from kirin import ir
from kirin.prelude import basic_no_opt
from kirin.emit.julia import EmitJulia


def emit(fn: ir.Method):
    with io.StringIO() as file:
        emit_ = EmitJulia(basic_no_opt, file)
        emit_.run(fn, tuple(fn.arg_names[1:])).expect()
        return file.getvalue()


def test_func():
    @basic_no_opt
    def emit_func(x: int, y: int):
        def foo():
            return x

        return foo

    generated = emit(emit_func)
    print(generated)
    assert "function emit_func(x::Int, y::Int)" in generated
    assert "@label block_0;" in generated
    assert "function foo()" in generated
    assert "@label block_1;" in generated
    assert "return x" in generated
    assert "return foo" in generated


def test_py_stmts():
    @basic_no_opt
    def emit_py_stmts(x: int, y: int):
        a1 = x * (x + y - x + 1)
        a2 = x == y
        a3 = x >= y
        a4 = x <= y
        a5 = x != y
        a6 = x > y
        a7 = x < y
        return a1, a2, a3, a4, a5, a6, a7

    generated = emit(emit_py_stmts)
    assert "var_0 = x + y" in generated
    assert "var_1 = var_0 - x" in generated
    assert "var_2 = var_1 + 1" in generated
    assert "a1 = x * var_2" in generated
    assert "a2 = x == y" in generated
    assert "a3 = x >= y" in generated
    assert "a4 = x <= y" in generated
    assert "a5 = x != y" in generated
    assert "a6 = x > y" in generated
    assert "a7 = x < y" in generated
    assert "var_3 = (a1, a2, a3, a4, a5, a6, a7)" in generated


def test_cf():
    @basic_no_opt
    def emit_cf(x: int, y: int):
        if x > y:
            return x
        else:
            return y

    emit_cf.print()
    generated = emit(emit_cf)

    assert "var_0 = x > y" in generated
    assert "if var_0" in generated
    assert "@goto block_1" in generated
    assert "@goto block_2" in generated



from kirin.ir import Method
from kirin.interp import Frame, MethodTable, ReturnValue, impl, concrete
from kirin.dialects.func.stmts import (
    Call,
    Invoke,
    Lambda,
    Return,
    GetField,
    ConstantNone,
)
from kirin.dialects.func.dialect import dialect


@dialect.register
class Interpreter(MethodTable):

    @impl(Call)
    def call(self, interp: concrete.Interpreter, frame: Frame, stmt: Call):
        mt: Method = frame.get(stmt.callee)
        _, result = interp.run_method(
            mt,
            interp.permute_values(
                mt.arg_names, frame.get_values(stmt.inputs), stmt.kwargs
            ),
        )
        return (result,)

    @impl(Invoke)
    def invoke(self, interp: concrete.Interpreter, frame: Frame, stmt: Invoke):
        _, result = interp.run_method(
            stmt.callee,
            interp.permute_values(
                stmt.callee.arg_names, frame.get_values(stmt.inputs), stmt.kwargs
            ),
        )
        return (result,)

    @impl(Return)
    def return_(self, interp: concrete.Interpreter, frame: Frame, stmt: Return):
        return ReturnValue(frame.get(stmt.value))

    @impl(ConstantNone)
    def const_none(
        self, interp: concrete.Interpreter, frame: Frame, stmt: ConstantNone
    ):
        return (None,)

    @impl(GetField)
    def getfield(self, interp: concrete.Interpreter, frame: Frame, stmt: GetField):
        mt: Method = frame.get(stmt.obj)
        return (mt.fields[stmt.field],)

    @impl(Lambda)
    def lambda_(self, interp: concrete.Interpreter, frame: Frame, stmt: Lambda):
        return (
            Method(
                mod=None,
                py_func=None,
                sym_name=stmt.name,
                arg_names=[
                    arg.name or str(idx)
                    for idx, arg in enumerate(stmt.body.blocks[0].args)
                ],
                dialects=interp.dialects,
                code=stmt,
                fields=frame.get_values(stmt.captured),
            ),
        )



from dataclasses import dataclass

from kirin.ir import Method, SSACFGRegion
from kirin.rewrite import (
    Walk,
    Chain,
    Fixpoint,
    WrapConst,
    Call2Invoke,
    ConstantFold,
    CFGCompactify,
    InlineGetItem,
    DeadCodeElimination,
)
from kirin.analysis import const
from kirin.passes.abc import Pass
from kirin.rewrite.abc import RewriteResult


@dataclass
class Fold(Pass):

    def unsafe_run(self, mt: Method) -> RewriteResult:
        constprop = const.Propagate(self.dialects)
        frame, _ = constprop.run_analysis(mt, no_raise=self.no_raise)
        result = Walk(WrapConst(frame)).rewrite(mt.code)
        result = (
            Fixpoint(
                Walk(
                    Chain(
                        ConstantFold(),
                        InlineGetItem(),
                        Call2Invoke(),
                        DeadCodeElimination(),
                    )
                )
            )
            .rewrite(mt.code)
            .join(result)
        )

        if mt.code.has_trait(SSACFGRegion):
            result = Walk(CFGCompactify()).rewrite(mt.code).join(result)

        return Fixpoint(Walk(DeadCodeElimination())).rewrite(mt.code).join(result)



from typing import Iterable
from dataclasses import field, dataclass

from kirin import ir
from kirin.print import Printable
from kirin.dialects import func
from kirin.print.printer import Printer


@dataclass
class CallGraph(Printable):
    """Call graph for a given [`ir.Method`][kirin.ir.Method].

    This class implements the [`kirin.graph.Graph`][kirin.graph.Graph] protocol.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    defs: dict[str, ir.Method] = field(default_factory=dict)
    """Mapping from symbol names to methods."""
    backedges: dict[str, set[str]] = field(default_factory=dict)
    """Mapping from symbol names to backedges."""

    def __init__(self, mt: ir.Method):
        self.defs = {}
        self.backedges = {}
        self.__build(mt)

    def __build(self, mt: ir.Method):
        self.defs[mt.sym_name] = mt
        for stmt in mt.callable_region.walk():
            if isinstance(stmt, func.Invoke):
                backedges = self.backedges.setdefault(stmt.callee.sym_name, set())
                backedges.add(mt.sym_name)
                self.__build(stmt.callee)

    def get_neighbors(self, node: str) -> Iterable[str]:
        """Get the neighbors of a node in the call graph."""
        return self.backedges.get(node, ())

    def get_edges(self) -> Iterable[tuple[str, str]]:
        """Get the edges of the call graph."""
        for node, neighbors in self.backedges.items():
            for neighbor in neighbors:
                yield node, neighbor

    def get_nodes(self) -> Iterable[str]:
        """Get the nodes of the call graph."""
        return self.defs.keys()

    def print_impl(self, printer: Printer) -> None:
        for idx, (caller, callee) in enumerate(self.backedges.items()):
            printer.plain_print(caller)
            printer.plain_print(" -> ")
            printer.print_seq(
                callee, delim=", ", prefix="[", suffix="]", emit=printer.plain_print
            )
            if idx < len(self.backedges) - 1:
                printer.print_newline()



import ast
from dataclasses import dataclass

from rich.console import Console


@dataclass
class SourceInfo:
    lineno: int
    col_offset: int
    end_lineno: int | None
    end_col_offset: int | None

    @classmethod
    def from_ast(cls, node: ast.AST, lineno_offset: int = 0, col_offset: int = 0):
        end_lineno = getattr(node, "end_lineno", None)
        end_col_offset = getattr(node, "end_col_offset", None)
        return cls(
            getattr(node, "lineno", 0) + lineno_offset,
            getattr(node, "col_offset", 0) + col_offset,
            end_lineno + lineno_offset if end_lineno is not None else None,
            end_col_offset + col_offset if end_col_offset is not None else None,
        )

    def offset(self, lineno_offset: int = 0, col_offset: int = 0):
        """Offset the source info by the given offsets.

        Args:
            lineno_offset (int): The line number offset.
            col_offset (int): The column offset.
        """
        self.lineno += lineno_offset
        self.col_offset += col_offset
        if self.end_lineno is not None:
            self.end_lineno += lineno_offset
        if self.end_col_offset is not None:
            self.end_col_offset += col_offset
        return self

    def error_hint(
        self,
        lines: list[str],
        err: Exception,
        *,
        file: str | None = None,
        indent: int = 2,
        show_lineno: bool = True,
        max_lines: int = 3,
        lineno_offset: int = 0,
    ) -> str:
        """Generate a hint for the error.

        Args:
            lines (list[str]): The lines of code.
            err (Exception): The error to display. If the error object has a
                `help` attribute, it will be used as the help message at the
                location of the error.
            file (str | None): The name of the file.
            indent (int): The indentation level.
            show_lineno (bool): Whether to show the line number.
            max_lines (int): The maximum number of lines to display.
            lineno_offset (int): The offset for the line number.

        Returns:
            str: The hint for the error.
        """
        help = getattr(err, "help", None)
        begin = max(0, self.lineno - max_lines - lineno_offset)
        end = max(
            max(self.lineno + max_lines, self.end_lineno or 0) - lineno_offset,
            0,
        )
        end = min(len(lines), end)  # make sure end is within bounds
        lines = lines[begin:end]
        error_lineno = self.lineno - lineno_offset - 1
        error_lineno_len = len(str(self.lineno))
        code_indent = min(map(self.__get_indent, lines), default=0)

        console = Console(force_terminal=True)
        with console.capture() as capture:
            console.print()
            console.print(
                f"File: [dim]{file or 'stdin'}:{self.lineno}[/dim]",
                markup=True,
                highlight=False,
            )
            emsg = "\n  ".join(err.args)
            console.print(f"[red]  {type(err).__name__}: {emsg}[/red]")
            for lineno, line in enumerate(lines, begin):
                line = " " * indent + line[code_indent:]
                if show_lineno:
                    if lineno == error_lineno:
                        line = f"{self.lineno}[dim]│[/dim]" + line
                    else:
                        line = "[dim]" + " " * (error_lineno_len) + "│[/dim]" + line
                console.print("  " + line, markup=True, highlight=False)
                if lineno == error_lineno:
                    console.print(
                        "  "
                        + self.__arrow(
                            code_indent, error_lineno_len, help, indent, show_lineno
                        ),
                        markup=True,
                        highlight=False,
                    )

            if end == error_lineno:
                console.print(
                    "  "
                    + self.__arrow(
                        code_indent, error_lineno_len, help, indent, show_lineno
                    ),
                    markup=True,
                    highlight=False,
                )

        return capture.get()

    def __arrow(
        self,
        code_indent: int,
        error_lineno_len: int,
        help,
        indent: int,
        show_lineno: bool,
    ) -> str:
        ret = " " * (self.col_offset - code_indent)
        if self.end_col_offset:
            ret += "^" * (self.end_col_offset - self.col_offset)
        else:
            ret += "^"

        ret = " " * indent + "[red]" + ret
        if help:
            ret += " help: " + str(help) + "[/red]"
        if show_lineno:
            ret = " " * error_lineno_len + "[dim]│[/dim]" + ret
        return ret

    @staticmethod
    def __get_indent(line: str) -> int:
        if len(line) == 0:
            return int(1e9)  # very large number
        return len(line) - len(line.lstrip())



"""Assignment dialect for Python.

This module contains the dialect for the Python assignment statement, including:

- Statements: `Alias`, `SetItem`.
- The lowering pass for the assignments.
- The concrete implementation of the assignment statements.

This dialects maps Python assignment syntax.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.print import Printer

dialect = ir.Dialect("py.assign")

T = types.TypeVar("T")


@statement(dialect=dialect)
class Alias(ir.Statement):
    name = "alias"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(T)
    target: ir.PyAttr[str] = info.attribute()
    result: ir.ResultValue = info.result(T)

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        with printer.rich(style="symbol"):
            printer.plain_print(self.target.data)

        with printer.rich(style="keyword"):
            printer.plain_print(" = ")

        printer.print(self.value)


@statement(dialect=dialect)
class SetItem(ir.Statement):
    name = "setitem"
    traits = frozenset({lowering.FromPythonCall()})
    obj: ir.SSAValue = info.argument(print=False)
    value: ir.SSAValue = info.argument(print=False)
    index: ir.SSAValue = info.argument(print=False)


@statement(dialect=dialect)
class SetAttribute(ir.Statement):
    name = "setattr"
    traits = frozenset({lowering.FromPythonCall()})
    obj: ir.SSAValue = info.argument(print=False)
    attr: str = info.attribute()
    value: ir.SSAValue = info.argument(print=False)


@statement(dialect=dialect)
class TypeAssert(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    got: ir.SSAValue = info.argument(print=False)
    expected: types.TypeAttribute = info.attribute()
    result: ir.ResultValue = info.result()

    def __init__(self, got: ir.SSAValue, *, expected: types.TypeAttribute):
        super().__init__(
            args=(got,),
            attributes={"expected": expected},
            result_types=(expected,),
            args_slice={"got": 0},
        )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Alias)
    def alias(self, interp, frame: interp.Frame, stmt: Alias):
        return (frame.get(stmt.value),)

    @interp.impl(SetItem)
    def setindex(self, interp, frame: interp.Frame, stmt: SetItem):
        frame.get(stmt.obj)[frame.get(stmt.index)] = frame.get(stmt.value)

    @interp.impl(SetAttribute)
    def set_attribute(self, interp, frame: interp.Frame, stmt: SetAttribute):
        obj = frame.get(stmt.obj)
        value = frame.get(stmt.value)
        setattr(obj, stmt.attr, value)

    # NOTE: we don't do much runtime type checking here, object with generic
    # types will unlikely work here.
    # TODO: consider runtime type checking by boxing the value
    @interp.impl(TypeAssert)
    def type_assert(self, interp_, frame: interp.Frame, stmt: TypeAssert):
        got = frame.get(stmt.got)
        got_type = types.PyClass(type(got))
        if not got_type.is_subseteq(stmt.expected):
            raise interp.WrapException(
                TypeError(f"Expected {stmt.expected}, got {got_type}")
            )
        return (frame.get(stmt.got),)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):
    @interp.impl(TypeAssert)
    def type_assert(
        self, interp_, frame: interp.Frame[types.TypeAttribute], stmt: TypeAssert
    ):
        got = frame.get(stmt.got)
        if got.is_subseteq(stmt.expected):
            return (got.meet(stmt.expected),)
        return (types.Bottom,)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Assign(self, state: lowering.State, node: ast.Assign) -> lowering.Result:
        result = state.lower(node.value)
        current_frame = state.current_frame
        match node:
            case ast.Assign(
                targets=[ast.Name(lhs_name, ast.Store())], value=ast.Name(_, ast.Load())
            ):
                stmt = Alias(
                    value=result.data[0], target=ir.PyAttr(lhs_name)
                )  # NOTE: this is guaranteed to be one result
                stmt.result.name = lhs_name
                current_frame.defs[lhs_name] = current_frame.push(stmt).result
            case _:
                for target in node.targets:
                    self.assign_item(state, target, result)

    def lower_AnnAssign(
        self, state: lowering.State, node: ast.AnnAssign
    ) -> lowering.Result:
        type_hint = self.get_hint(state, node.annotation)
        value = state.lower(node.value).expect_one()
        stmt = state.current_frame.push(TypeAssert(got=value, expected=type_hint))
        self.assign_item_value(state, node.target, stmt.result)

    def lower_AugAssign(
        self, state: lowering.State, node: ast.AugAssign
    ) -> lowering.Result:
        match node.target:
            case ast.Name(name, ast.Store()):
                rhs = ast.Name(name, ast.Load())
            case ast.Attribute(obj, attr, ast.Store()):
                rhs = ast.Attribute(obj, attr, ast.Load())
            case ast.Subscript(obj, slice, ast.Store()):
                rhs = ast.Subscript(obj, slice, ast.Load())
            case _:
                raise lowering.BuildError(f"unsupported target {node.target}")
        self.assign_item_value(
            state,
            node.target,
            state.lower(ast.BinOp(rhs, node.op, node.value)).expect_one(),
        )

    def lower_NamedExpr(
        self, state: lowering.State, node: ast.NamedExpr
    ) -> lowering.Result:
        value = state.lower(node.value).expect_one()
        self.assign_item_value(state, node.target, value)
        return value

    @classmethod
    def assign_item_value(cls, state: lowering.State, target, value: ir.SSAValue):
        current_frame = state.current_frame
        match target:
            case ast.Name(name, ast.Store()):
                value.name = name
                current_frame.defs[name] = value
            case ast.Attribute(obj, attr, ast.Store()):
                obj = state.lower(obj).expect_one()
                stmt = SetAttribute(obj, value, attr=attr)
                current_frame.push(stmt)
            case ast.Subscript(obj, slice, ast.Store()):
                obj = state.lower(obj).expect_one()
                slice = state.lower(slice).expect_one()
                stmt = SetItem(obj=obj, index=slice, value=value)
                current_frame.push(stmt)
            case _:
                raise lowering.BuildError(f"unsupported target {target}")

    @classmethod
    def assign_item(cls, state: lowering.State, target, result: lowering.State.Result):
        match target:
            case ast.Tuple(elts, ast.Store()):
                if len(elts) != len(result.data):
                    raise lowering.BuildError(
                        f"tuple assignment length mismatch: {len(elts)} != {len(result.data)}"
                    )
                for target, value in zip(elts, result.data):
                    cls.assign_item_value(state, target, value)
            case _:
                cls.assign_item_value(state, target, result.expect_one())



import pytest

from kirin.ir import Block
from kirin.dialects import py


def test_stmt():
    a = py.Constant(0)
    x = py.Constant(1)
    y = py.Constant(2)
    z = py.Add(lhs=x.result, rhs=y.result)

    bb1 = Block([a, x, y, z])
    assert bb1.first_stmt == a
    bb1.print()

    a.delete()
    assert bb1.first_stmt == x
    bb1.print()

    a.insert_before(x)
    bb1.print()
    assert bb1.first_stmt == a

    a.delete()
    a.insert_after(x)
    bb1.stmts.at(1) == a  # type: ignore

    with pytest.raises(ValueError):
        a.insert_after(x)

    with pytest.raises(ValueError):
        a.insert_before(x)


def test_block_append():
    block = Block()
    block.stmts.append(py.Constant(1))
    block.stmts.append(py.Constant(1))
    block.print()
    assert len(block.stmts) == 2


def test_stmt_from_stmt():

    x = py.Constant(1)

    x.result.hints["const"] = py.constant.types.Int

    y = x.from_stmt(x)

    assert y.result.hints["const"] == py.constant.types.Int



from kirin.prelude import basic_no_opt
from kirin.analysis.callgraph import CallGraph


@basic_no_opt
def abc(a, b):
    return a + b


@basic_no_opt
def bcd(a, b):
    return a - b


@basic_no_opt
def cde(a, b):
    return abc(a, b) + bcd(a, b)


@basic_no_opt
def defg(a, b):
    return cde(a, b) + abc(a, b)


@basic_no_opt
def efg(a, b):
    return defg(a, b) + bcd(a, b)


def test_callgraph():
    graph = CallGraph(efg)
    graph.print()
    assert "cde" in graph.get_neighbors("abc")
    assert "defg" in graph.get_neighbors("abc")
    assert "cde" in graph.get_neighbors("abc")
    assert "defg" in graph.get_neighbors("abc")



from dataclasses import field, dataclass

from kirin.passes import Pass
from kirin.rewrite import (
    Walk,
    Chain,
    Inline,
    Fixpoint,
    WrapConst,
    Call2Invoke,
    ConstantFold,
    CFGCompactify,
    InlineGetItem,
    InlineGetField,
    DeadCodeElimination,
)
from kirin.analysis import const
from kirin.ir.method import Method
from kirin.rewrite.abc import RewriteResult


@dataclass
class Fold(Pass):
    constprop: const.Propagate = field(init=False)

    def __post_init__(self):
        self.constprop = const.Propagate(self.dialects)

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = RewriteResult()
        frame, _ = self.constprop.run_analysis(mt, no_raise=self.no_raise)
        result = Walk(WrapConst(frame)).rewrite(mt.code).join(result)
        rule = Chain(
            ConstantFold(),
            Call2Invoke(),
            InlineGetField(),
            InlineGetItem(),
            DeadCodeElimination(),
        )
        result = Fixpoint(Walk(rule)).rewrite(mt.code).join(result)
        result = Walk(Inline(lambda _: True)).rewrite(mt.code).join(result)
        result = Fixpoint(CFGCompactify()).rewrite(mt.code).join(result)
        return result



"""A Python-like structural Control Flow dialect.

This dialect provides constructs for expressing control flow in a structured
manner. The dialect provides constructs for expressing loops and conditionals.
Unlike MLIR SCF dialect, this dialect does not restrict the control flow to
statically analyzable forms. This dialect is designed to be compatible with
Python native control flow constructs.

This dialect depends on the following dialects:
- `eltype`: for obtaining the element type of a value.
"""

from . import (
    trim as trim,
    absint as absint,
    interp as interp,
    unroll as unroll,
    lowering as lowering,
    constprop as constprop,
    typeinfer as typeinfer,
)
from .stmts import For as For, Yield as Yield, IfElse as IfElse
from ._dialect import dialect as dialect



from kirin import ir
from kirin.prelude import python_basic
from kirin.dialects import scf, func, lowering

# TODO:
# test_cons
# cond = py.Constant(True)
# then_body = ir.Region(ir.Block())
# else_body = ir.Region(ir.Block())

# ifelse = scf.IfElse(cond.result, ir.Block([scf.Yield(cond.result)]), ir.Block([scf.Yield(cond.result)]))
# ifelse.print()

# ir.Block([ifelse]).print()


@ir.dialect_group(python_basic.union([func, scf, lowering.func]))
def kernel(self):
    def run_pass(method):
        pass

    return run_pass


@kernel
def main(x):
    if x > 0:
        y = x + 1
        z = y + 1
        return z
    else:
        y = x + 2
        z = y + 2

    if x < 0:
        y = y + 3
        z = y + 3
    else:
        y = x + 4
        z = y + 4
    return y, z


main.print()
# print(main(1))



"""This module contains the dialects for choosing different lowering strategies.

The dialects defined inside this module do not provide any new statements, it only
provide different lowering strategies for existing statements.
"""

from . import cf as cf, call as call, func as func, range as range



# NOTE: this module is only interface, will be used inside
# the `ir` module try to minimize the dependencies as much
# as possible

from __future__ import annotations

import ast
import inspect
from abc import ABC
from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeAlias
from dataclasses import dataclass

from kirin.ir.attrs import types
from kirin.lowering.abc import Result
from kirin.lowering.exception import BuildError
from kirin.lowering.python.glob import GlobalEvalError

if TYPE_CHECKING:
    from kirin.lowering.state import State


LoweringTransform: TypeAlias = Callable[[Any, "State[ast.AST]", ast.Call], Result]


@dataclass
class Transform:
    objs: tuple[Callable, ...]
    func: LoweringTransform


@dataclass
class akin:
    obj: Callable

    def __call__(
        self,
        func: LoweringTransform | Transform,
    ) -> Transform:
        if isinstance(func, Transform):
            return Transform((self.obj,) + func.objs, func.func)

        if not func.__name__.startswith("lower_Call_"):
            raise SyntaxError(
                "lowering function should be prefixed with lower_Call_"
                f" but got {func.__name__}"
            )
        return Transform((self.obj,), func)


@dataclass
class FromPythonAST(ABC):
    callee_table: ClassVar[dict[object, Transform]]
    """a table of lowering transforms for ast.Call based
    on the callable object if avaiable as a global value.
    """

    def __init_subclass__(cls) -> None:
        # init the subclass first
        super().__init_subclass__()
        cls.callee_table = {}
        for _, value in inspect.getmembers(cls):
            if isinstance(value, Transform):
                for obj in value.objs:
                    cls.callee_table[obj] = value

    @property
    def names(self) -> list[str]:  # show the name without lower_
        return [name[6:] for name in dir(self) if name.startswith("lower_")]

    def lower(self, state: State[ast.AST], node: ast.AST) -> Result:
        """Entry point of dialect specific lowering."""
        return getattr(self, f"lower_{node.__class__.__name__}", self.unreachable)(
            state, node
        )

    def unreachable(self, state: State[ast.AST], node: ast.AST) -> Result:
        raise BuildError(f"unreachable reached for {node.__class__.__name__}")

    @staticmethod
    def get_hint(state: State[ast.AST], node: ast.expr | None) -> types.TypeAttribute:
        if node is None:
            return types.AnyType()

        try:
            t = state.get_global(node).data
            return types.hint2type(t)
        except GlobalEvalError as e:
            state.source = e.source.offset(state.lineno_offset, state.col_offset)
            raise e
        except Exception as e:  # noqa: E722
            raise BuildError(f"expect a type hint, got {ast.unparse(node)}") from e


class NoSpecialLowering(FromPythonAST):
    pass



from abc import ABC
from typing import TypeVar
from dataclasses import field, dataclass

from kirin import ir, interp
from kirin.worklist import WorkList

ValueType = TypeVar("ValueType")


@dataclass
class EmitFrame(interp.Frame[ValueType]):
    worklist: WorkList[interp.Successor] = field(default_factory=WorkList)
    block_ref: dict[ir.Block, ValueType] = field(default_factory=dict)


FrameType = TypeVar("FrameType", bound=EmitFrame)


@dataclass
class EmitABC(interp.BaseInterpreter[FrameType, ValueType], ABC):

    def run_callable_region(
        self, frame: FrameType, code: ir.Statement, region: ir.Region
    ) -> ValueType:
        results = self.eval_stmt(frame, code)
        if isinstance(results, tuple):
            if len(results) == 0:
                return self.void
            elif len(results) == 1:
                return results[0]
        raise interp.InterpreterError(f"Unexpected results {results}")

    def run_ssacfg_region(
        self, frame: FrameType, region: ir.Region
    ) -> tuple[ValueType, ...]:
        frame.worklist.append(
            interp.Successor(region.blocks[0], frame.get_values(region.blocks[0].args))
        )
        while (succ := frame.worklist.pop()) is not None:
            block_header = self.emit_block(frame, succ.block)
            frame.block_ref[succ.block] = block_header
        return ()

    def emit_attribute(self, attr: ir.Attribute) -> ValueType:
        return getattr(
            self, f"emit_type_{type(attr).__name__}", self.emit_attribute_fallback
        )(attr)

    def emit_attribute_fallback(self, attr: ir.Attribute) -> ValueType:
        if (method := self.registry.attributes.get(type(attr))) is not None:
            return method(self, attr)
        raise NotImplementedError(f"Attribute {type(attr)} not implemented")

    def emit_stmt_begin(self, frame: FrameType, stmt: ir.Statement) -> None:
        return

    def emit_stmt_end(self, frame: FrameType, stmt: ir.Statement) -> None:
        return

    def emit_block_begin(self, frame: FrameType, block: ir.Block) -> None:
        return

    def emit_block_end(self, frame: FrameType, block: ir.Block) -> None:
        return

    def emit_block(self, frame: FrameType, block: ir.Block) -> ValueType:
        self.emit_block_begin(frame, block)
        stmt = block.first_stmt
        while stmt is not None:
            if self.consume_fuel() == self.FuelResult.Stop:
                raise interp.FuelExhaustedError("fuel exhausted")

            self.emit_stmt_begin(frame, stmt)
            stmt_results = self.eval_stmt(frame, stmt)
            self.emit_stmt_end(frame, stmt)

            match stmt_results:
                case tuple(values):
                    frame.set_values(stmt._results, values)
                case interp.ReturnValue(_) | interp.YieldValue(_):
                    pass
                case _:
                    raise ValueError(f"Unexpected result {stmt_results}")

            stmt = stmt.next_stmt

        self.emit_block_end(frame, block)
        return frame.block_ref[block]



from dataclasses import field, dataclass

from kirin import ir, types, interp
from kirin.analysis.forward import ForwardExtra, ForwardFrame

from .lattice import Value, Result, Unknown


@dataclass
class Frame(ForwardFrame[Result]):
    should_be_pure: set[ir.Statement] = field(default_factory=set)
    """If any ir.MaybePure is actually pure."""
    frame_is_not_pure: bool = False
    """If we hit any non-pure statement."""


@dataclass
class Propagate(ForwardExtra[Frame, Result]):
    """Forward dataflow analysis for constant propagation.

    This analysis is a forward dataflow analysis that propagates constant values
    through the program. It uses the `Result` lattice to track the constant
    values and purity of the values.

    The analysis is implemented as a forward dataflow analysis, where the
    `eval_stmt` method is overridden to handle the different types of statements
    in the IR. The analysis uses the `interp.Interpreter` to evaluate the
    statements and propagate the constant values.

    When a statement is registered under the "constprop" key in the method table,
    the analysis will call the method to evaluate the statement instead of using
    the interpreter. This allows for custom handling of statements.
    """

    keys = ["constprop"]
    lattice = Result

    _interp: interp.Interpreter = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self._interp = interp.Interpreter(
            self.dialects,
            fuel=self.fuel,
            debug=self.debug,
            max_depth=self.max_depth,
            max_python_recursion_depth=self.max_python_recursion_depth,
        )

    def initialize(self):
        super().initialize()
        self._interp.initialize()
        return self

    def new_frame(self, code: ir.Statement) -> Frame:
        return Frame.from_func_like(code)

    def try_eval_const_pure(
        self,
        frame: Frame,
        stmt: ir.Statement,
        values: tuple[Value, ...],
    ) -> interp.StatementResult[Result]:
        _frame = self._interp.new_frame(frame.code)
        _frame.set_values(stmt.args, tuple(x.data for x in values))
        method = self._interp.lookup_registry(frame, stmt)
        if method is not None:
            value = method(self._interp, _frame, stmt)
        else:
            return (Unknown(),)
        match value:
            case tuple():
                return tuple(Value(each) for each in value)
            case interp.ReturnValue(ret):
                return interp.ReturnValue(Value(ret))
            case interp.YieldValue(yields):
                return interp.YieldValue(tuple(Value(each) for each in yields))
            case interp.Successor(block, args):
                return interp.Successor(
                    block,
                    *tuple(Value(each) for each in args),
                )

    def eval_stmt(
        self, frame: Frame, stmt: ir.Statement
    ) -> interp.StatementResult[Result]:
        method = self.lookup_registry(frame, stmt)
        if method is None:
            if stmt.has_trait(ir.ConstantLike):
                return self.try_eval_const_pure(frame, stmt, ())
            elif stmt.has_trait(ir.Pure):
                values = frame.get_values(stmt.args)
                if types.is_tuple_of(values, Value):
                    return self.try_eval_const_pure(frame, stmt, values)

            if stmt.has_trait(ir.Pure):
                return (Unknown(),)  # no implementation but pure
            # not pure, and no implementation, let's say it's not pure
            frame.frame_is_not_pure = True
            return (Unknown(),)

        ret = method(self, frame, stmt)
        if stmt.has_trait(ir.IsTerminator) or stmt.has_trait(ir.Pure):
            return ret
        elif not stmt.has_trait(ir.MaybePure):  # cannot be pure at all
            frame.frame_is_not_pure = True
        elif (
            stmt not in frame.should_be_pure
        ):  # implementation cannot decide if it's pure
            frame.frame_is_not_pure = True
        return ret

    def run_method(
        self, method: ir.Method, args: tuple[Result, ...]
    ) -> tuple[Frame, Result]:
        return self.run_callable(method.code, (Value(method),) + args)



from kirin import interp

from . import stmts
from ._dialect import dialect


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(stmts.UAdd)
    def uadd(self, interp, frame: interp.Frame, stmt: stmts.UAdd):
        return (+frame.get(stmt.value),)

    @interp.impl(stmts.USub)
    def usub(self, interp, frame: interp.Frame, stmt: stmts.USub):
        return (-frame.get(stmt.value),)

    @interp.impl(stmts.Not)
    def not_(self, interp, frame: interp.Frame, stmt: stmts.Not):
        return (not frame.get(stmt.value),)

    @interp.impl(stmts.Invert)
    def invert(self, interp, frame: interp.Frame, stmt: stmts.Invert):
        return (~frame.get(stmt.value),)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Optional,
    Sequence,
    Generator,
    overload,
)
from contextlib import contextmanager
from dataclasses import field, dataclass

from kirin.ir import Block, Region, SSAValue, Statement
from kirin.source import SourceInfo

from .abc import ASTNodeType
from .frame import Frame, CallbackFn
from .stream import StmtStream
from .exception import BuildError

if TYPE_CHECKING:
    from .abc import LoweringABC


@dataclass
class State(Generic[ASTNodeType]):
    """State of the lowering process.

    This class is used to store the state of the lowering process.
    It contains the current frame, the current block, and
    the current source.
    """

    parent: LoweringABC[ASTNodeType]
    """the parent lowering transform"""
    source: SourceInfo | None = field(default=None, kw_only=True)
    "source info of the current node"
    lines: list[str] = field(default_factory=list, kw_only=True)
    "source lines of the code being lowered"
    file: str | None = field(default=None, kw_only=True)
    "file name of the current node"
    lineno_offset: int = field(default=0, kw_only=True)
    "lineno offset at the beginning of the source"
    col_offset: int = field(default=0, kw_only=True)
    "column offset at the beginning of the source"
    _current_frame: Frame | None = field(default=None, init=False, repr=False)
    "current frame being lowered"

    def __repr__(self) -> str:
        return f"lowering.State(current_frame={self._current_frame})"

    @property
    def code(self):
        """Obtain the code generated by the lowering process."""
        region = self.root_frame.curr_region
        if not region.blocks:
            raise ValueError("No block generated")

        stmt = region.blocks[0].first_stmt
        if stmt is None:
            raise ValueError("No code generated")
        return stmt

    @property
    def root_frame(self) -> Frame[ASTNodeType]:
        """root frame of the lowering process"""
        if self._current_frame is None:
            raise ValueError("current frame is None")
        root = self._current_frame
        while root.parent is not None:
            root = root.parent
        return root

    @property
    def current_frame(self) -> Frame[ASTNodeType]:
        """current frame being lowered"""
        if self._current_frame is None:
            raise ValueError("current frame is None")
        return self._current_frame

    @dataclass
    class Result:
        """A proxy object to the result of the lowering process.

        Use `.data` to access the result of the lowering process.
        Use `.expect_one()` to assert that the result is a single value.
        """

        data: tuple[SSAValue, ...]

        def expect_one(self) -> SSAValue:
            if len(self.data) == 1:
                return self.data[0]
            raise BuildError("expected a value, but got None")

    def lower(self, node: ASTNodeType):
        result = self.parent.visit(self, node)
        if isinstance(result, Statement):
            return self.Result(tuple(result._results))
        elif result is None:
            return self.Result(tuple())
        elif isinstance(result, SSAValue):
            return self.Result((result,))
        return self.Result(result)

    def get_literal(self, value) -> SSAValue:
        return self.parent.lower_literal(self, value)

    @overload
    def get_global(
        self, node: ASTNodeType, *, no_raise: Literal[True] | bool
    ) -> LoweringABC.Result | None: ...

    @overload
    def get_global(
        self, node: ASTNodeType, *, no_raise: Literal[False] = False
    ) -> LoweringABC.Result: ...

    def get_global(self, node: ASTNodeType, *, no_raise: bool = False):
        """Get the global value of a node.
        Args:
            node (ASTNodeType): the node to get the global value of.
            no_raise (bool): if True, do not raise an exception if the value is not found.
        Returns:
            `LoweringABC.Result`: a proxy object to the global value. `.data` is the
                value, and `.expect(type)` will raise an exception if the value is the expected type.
        """
        if no_raise:
            return self.parent.lower_global_no_raise(self, node)
        return self.parent.lower_global(self, node)

    def push_frame(self, frame: Frame):
        assert frame is not self._current_frame, "Frame already pushed"
        frame.parent = self._current_frame
        self._current_frame = frame
        return frame

    def pop_frame(self, finalize_next: bool = True):
        """Pop the current frame and return it.

        Args:
            finalize_next(bool): If True, append the next block of the current frame.

        Returns:
            Frame: The popped frame.
        """
        if self._current_frame is None:
            raise ValueError("No frame to pop")
        frame = self._current_frame

        if finalize_next and frame.next_block.parent is None:
            frame.push(frame.next_block)
        self._current_frame = frame.parent
        return frame

    @contextmanager
    def frame(
        self,
        stmts: Sequence[ASTNodeType] | StmtStream[ASTNodeType],
        region: Optional[Region] = None,
        entr_block: Optional[Block] = None,
        next_block: Optional[Block] = None,
        globals: dict[str, Any] | None = None,
        capture_callback: Optional[CallbackFn] = None,
        finalize_next: bool = True,
    ) -> Generator[Frame[ASTNodeType], Any, None]:
        """Context manager to push a new frame and pop it after the block."""

        if not isinstance(stmts, StmtStream):
            stmts = StmtStream(stmts)

        region = region or Region()
        region.source = self.source

        entr_block = entr_block or Block()
        region.blocks.append(entr_block)

        if self._current_frame is not None:
            globals = globals or self.current_frame.globals
        else:
            globals = globals or {}

        frame = Frame(
            state=self,
            stream=stmts,
            curr_region=region or Region(entr_block),
            entr_block=entr_block,
            curr_block=entr_block,
            next_block=next_block or Block(),
            globals=globals,
            capture_callback=capture_callback,
        )
        self.push_frame(frame)
        try:
            yield frame
        finally:
            self.pop_frame(finalize_next)

    def error_hint(
        self,
        e: Exception,
        *,
        max_lines: int = 3,
        indent: int = 0,
        show_lineno: bool = True,
    ) -> str:
        """Generate an error hint for the given error.
        Args:
            e (Exception): The error to generate a hint for. If the error object
                has a `help` attribute, it will be used as the help message at the
                location of the error.
            max_lines (int): The maximum number of lines to show in the hint.
            indent (int): The indentation level for the hint.
            show_lineno (bool): Whether to show the line number in the hint.
        Returns:
            str: The generated error hint.
        """
        if self.source is None:
            return str(e)

        return self.source.error_hint(
            self.lines,
            e,
            file=self.file,
            indent=indent,
            show_lineno=show_lineno,
            max_lines=max_lines,
            lineno_offset=self.lineno_offset,
        )



from __future__ import annotations

import ast
import inspect
import textwrap
from types import ModuleType
from typing import Any, Callable, Iterable
from dataclasses import dataclass

from kirin import ir
from kirin.source import SourceInfo
from kirin.registry import LoweringRegistry
from kirin.lowering.abc import Result, LoweringABC
from kirin.lowering.state import State
from kirin.lowering.exception import BuildError

from .glob import GlobalExprEval
from .traits import FromPythonCall, FromPythonWithSingleItem
from .binding import Binding


@dataclass
class Python(LoweringABC[ast.AST]):
    """Python lowering transform.

    This class is used to lower Python AST nodes to IR statements via
    a visitor pattern.

    !!! note
        the visitor pattern is not using the `ast.NodeVisitor` class
        because it customize the visit method to pass the lowering state
        and the source information to the visitor methods.
    """

    registry: LoweringRegistry
    max_lines: int = 3
    hint_indent: int = 2
    hint_show_lineno: bool = True
    stacktrace: bool = False
    """If True, print the stacktrace of the error."""

    def __init__(
        self,
        dialects: ir.DialectGroup | Iterable[ir.Dialect | ModuleType],
        *,
        keys: list[str] | None = None,
        max_lines: int = 3,
        hint_indent: int = 2,
        hint_lineno: bool = True,
        stacktrace: bool = False,
    ):
        if isinstance(dialects, ir.DialectGroup):
            self.dialects = dialects
        else:
            self.dialects = ir.DialectGroup(dialects)

        self.max_lines = max_lines
        self.registry = self.dialects.registry.ast(keys=keys or ["main", "default"])
        self.hint_indent = hint_indent
        self.hint_show_lineno = hint_lineno
        self.stacktrace = stacktrace

    def python_function(
        self,
        func: Callable,
        *,
        globals: dict[str, Any] | None = None,
        lineno_offset: int = 0,
        col_offset: int = 0,
        compactify: bool = True,
    ):
        file = inspect.getfile(func)
        source = textwrap.dedent(inspect.getsource(func))
        if globals:
            globals.update(func.__globals__)
        else:
            globals = func.__globals__

        try:
            nonlocals = inspect.getclosurevars(func).nonlocals
        except Exception:
            nonlocals = {}
        globals.update(nonlocals)
        region = self.run(
            ast.parse(source).body[0],
            source=source,
            globals=globals,
            file=file,
            lineno_offset=lineno_offset,
            col_offset=col_offset,
            compactify=compactify,
        )
        if not region.blocks:
            raise ValueError("No block generated")

        code = region.blocks[0].first_stmt
        if code is None:
            raise ValueError("No code generated")
        return code

    def run(
        self,
        stmt: ast.AST,
        *,
        source: str | None = None,
        globals: dict[str, Any] | None = None,
        file: str | None = None,
        lineno_offset: int = 0,
        col_offset: int = 0,
        compactify: bool = True,
    ) -> ir.Region:
        source = source or ast.unparse(stmt)
        state = State(
            self,
            source=SourceInfo.from_ast(stmt, lineno_offset, col_offset),
            file=file,
            lines=source.splitlines(),
            lineno_offset=lineno_offset,
            col_offset=col_offset,
        )

        with state.frame([stmt], globals=globals) as frame:
            try:
                self.visit(state, stmt)
            except BuildError as e:
                hint = state.error_hint(
                    e,
                    max_lines=self.max_lines,
                    indent=self.hint_indent,
                    show_lineno=self.hint_show_lineno,
                )
                if self.stacktrace:
                    raise Exception(
                        f"{e.args[0]}\n\n{hint}",
                        *e.args[1:],
                    ) from e
                else:
                    e.args = (hint,)
                    raise e

            region = frame.curr_region

        if compactify:
            from kirin.rewrite import Walk, CFGCompactify

            Walk(CFGCompactify()).rewrite(region)
        return region

    def lower_literal(self, state: State[ast.AST], value) -> ir.SSAValue:
        return state.lower(ast.Constant(value=value)).expect_one()

    def lower_global(self, state: State[ast.AST], node: ast.AST) -> LoweringABC.Result:
        return LoweringABC.Result(GlobalExprEval(state.current_frame).visit(node))

    # Python AST visitor methods
    def visit(self, state: State[ast.AST], node: ast.AST) -> Result:
        if hasattr(node, "lineno"):
            state.source = SourceInfo.from_ast(
                node, state.lineno_offset, state.col_offset
            )
        name = node.__class__.__name__
        if name in self.registry.ast_table:
            return self.registry.ast_table[name].lower(state, node)
        return getattr(self, f"visit_{name}", self.generic_visit)(state, node)

    def generic_visit(self, state: State[ast.AST], node: ast.AST) -> Result:
        raise BuildError(f"Cannot lower {node.__class__.__name__} node: {node}")

    def visit_Call(self, state: State[ast.AST], node: ast.Call) -> Result:
        if hasattr(node.func, "lineno"):
            state.source = SourceInfo.from_ast(
                node.func, state.lineno_offset, state.col_offset
            )

        global_callee_result = state.get_global(node.func, no_raise=True)
        if global_callee_result is None:
            return self.visit_Call_local(state, node)

        global_callee = global_callee_result.data

        if isinstance(global_callee, Binding):
            global_callee = global_callee.parent

        if isinstance(global_callee, ir.Method):
            return self.visit_Call_Method(state, node, global_callee)
        elif inspect.isclass(global_callee) and issubclass(global_callee, ir.Statement):
            return self.visit_Call_Class_Statement(state, node, global_callee)
        else:
            return self.visit_Call_table(state, node, global_callee)

    def visit_Call_table(self, state: State[ast.AST], node: ast.Call, global_callee):
        if method := self.registry.callee_table.get(global_callee):
            return method(state, node)
        return self.visit_Call_generic(state, node, global_callee)

    def visit_Call_generic(self, state: State[ast.AST], node: ast.Call, global_callee):
        # symbol exist in global, but not ir.Statement, it may refer to a
        # local value that shadows the global value
        try:
            return self.visit_Call_local(state, node)
        except BuildError:
            # symbol exist in global, but not ir.Statement, not found in locals either
            # this means the symbol is referring to an external uncallable object
            # try to hint the user
            if inspect.isfunction(global_callee):
                raise BuildError(
                    f"unsupported callee: {repr(global_callee)}."
                    "Are you trying to call a python function? This is not supported."
                )
            else:  # well not much we can do, can't hint
                raise BuildError(
                    f"unsupported call to {repr(global_callee)}, "
                    "expected a kernel function (Method), "
                    "wrapped statement (Binding) or a supported Python function"
                    f", got {type(global_callee)}"
                )

    def visit_Call_Class_Statement(
        self, state: State[ast.AST], node: ast.Call, global_callee: type[ir.Statement]
    ):
        if global_callee.dialect is None:
            raise BuildError(f"unsupported dialect `None` for {global_callee.name}")

        if global_callee.dialect not in self.dialects.data:
            raise BuildError(f"unsupported dialect `{global_callee.dialect.name}`")

        if (trait := global_callee.get_trait(FromPythonCall)) is not None:
            return trait.lower(global_callee, state, node)
        raise BuildError(
            f"invalid call syntax for {global_callee.__name__}, "
            f"expected FromPythonCall trait to be implemented"
            f" for {global_callee.__name__}"
        )

    def visit_Call_Method(
        self, state: State[ast.AST], node: ast.Call, global_callee: ir.Method
    ) -> Result:
        if "Call_global_method" in self.registry.ast_table:
            return self.registry.ast_table[
                "Call_global_method"
            ].lower_Call_global_method(state, global_callee, node)
        raise BuildError("`lower_Call_global_method` not implemented")

    def visit_Call_local(self, state: State[ast.AST], node: ast.Call) -> Result:
        callee = state.lower(node.func).expect_one()
        if "Call_local" in self.registry.ast_table:
            return self.registry.ast_table["Call_local"].lower_Call_local(
                state, callee, node
            )
        raise BuildError("`lower_Call_local` not implemented")

    def visit_With(self, state: State[ast.AST], node: ast.With) -> Result:
        if len(node.items) != 1:
            raise BuildError("expected exactly one item in with statement")

        item = node.items[0]
        if not isinstance(item.context_expr, ast.Call):
            raise BuildError("expected context expression to be a call")

        global_callee = state.get_global(item.context_expr.func).data
        if isinstance(global_callee, Binding):
            global_callee = global_callee.parent

        if not issubclass(global_callee, ir.Statement):
            raise BuildError(
                f"expected context expression to be a statement, got {global_callee}"
            )

        if trait := global_callee.get_trait(FromPythonWithSingleItem):
            return trait.lower(global_callee, state, node)

        raise BuildError(
            f"invalid with syntax for {global_callee.__name__}, "
            f"expected FromPythonWithSingleItem trait"
            " to be implemented"
            f" for {global_callee.__name__}"
        )



from dataclasses import dataclass


# errors
class InterpreterError(Exception):
    """Generic interpreter error.

    This is the base class for all interpreter errors. Interpreter
    errors will be catched by the interpreter and handled appropriately
    as an error with stack trace (of Kirin, not Python) from the interpreter.
    """

    pass


@dataclass
class WrapException(InterpreterError):
    """A special interpreter error that wraps a Python exception."""

    exception: Exception


class FuelExhaustedError(InterpreterError):
    """An error raised when the interpreter runs out of fuel."""

    pass



from dataclasses import field, dataclass


@dataclass
class RewriteResult:
    terminated: bool = field(default=False, kw_only=True)
    has_done_something: bool = field(default=False, kw_only=True)
    exceeded_max_iter: bool = field(default=False, kw_only=True)

    def join(self, other: "RewriteResult") -> "RewriteResult":
        return RewriteResult(
            terminated=self.terminated or other.terminated,
            has_done_something=self.has_done_something or other.has_done_something,
            exceeded_max_iter=self.exceeded_max_iter or other.exceeded_max_iter,
        )



from kirin import ir, types
from kirin.analysis import const
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult
from kirin.dialects.py.constant import Constant

from ..stmts import IListType
from ..runtime import IList


class ConstList2IList(RewriteRule):
    """Rewrite type annotation for SSAValue with constant `IList`
    in `Hinted` type. This should be run after constant folding and
    `WrapConst` rule.
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, Constant):
            return self.rewrite_Constant(node)

        has_done_something = False
        for result in node.results:
            if not isinstance(hint := result.hints.get("const"), const.Value):
                continue

            typ = result.type
            data = hint.data
            if isinstance(typ, types.PyClass) and typ.is_subseteq(types.PyClass(IList)):
                has_done_something = self._rewrite_IList_type(result, data)
            elif isinstance(typ, types.Generic) and typ.body.is_subseteq(
                types.PyClass(IList)
            ):
                has_done_something = self._rewrite_IList_type(result, data)
        return RewriteResult(has_done_something=has_done_something)

    def rewrite_Constant(self, node: Constant) -> RewriteResult:
        if isinstance(node.value, ir.PyAttr) and isinstance(node.value.data, list):
            stmt = Constant(value=IList(data=node.value.data))
            node.replace_by(stmt)
            self._rewrite_IList_type(stmt.result, node.value.data)
            return RewriteResult(has_done_something=True)
        return RewriteResult()

    def _rewrite_IList_type(self, result: ir.SSAValue, data):
        if not isinstance(data, IList):
            return False

        if not data.data:
            return False

        elem_type = types.PyClass(type(data[0]))
        for elem in data.data[1:]:
            elem_type = elem_type.join(types.PyClass(type(elem)))

        result.type = IListType[elem_type, types.Literal(len(data.data))]
        result.hints["const"] = const.Value(data)
        return True



from .cse import CommonSubexpressionElimination as CommonSubexpressionElimination
from .dce import DeadCodeElimination as DeadCodeElimination
from .fold import ConstantFold as ConstantFold
from .walk import Walk as Walk
from .alias import InlineAlias as InlineAlias
from .chain import Chain as Chain
from .inline import Inline as Inline
from .getitem import InlineGetItem as InlineGetItem
from .fixpoint import Fixpoint as Fixpoint
from .getfield import InlineGetField as InlineGetField
from .apply_type import ApplyType as ApplyType
from .compactify import CFGCompactify as CFGCompactify
from .wrap_const import WrapConst as WrapConst
from .call2invoke import Call2Invoke as Call2Invoke
from .type_assert import InlineTypeAssert as InlineTypeAssert



"""This module contains custom exception handling
for the Kirin-based compilers.
"""

import sys
import types

stacktrace = False


class NoPythonStackTrace(Exception):
    pass


def enable_stracetrace():
    """Enable the stacktrace for all exceptions."""
    global stacktrace
    stacktrace = True


def disable_stracetrace():
    """Disable the stacktrace for all exceptions."""
    global stacktrace
    stacktrace = False


def exception_handler(exc_type, exc_value, exc_tb: types.TracebackType):
    """Custom exception handler to format and print exceptions."""
    if not stacktrace and issubclass(exc_type, NoPythonStackTrace):
        print(exc_value, file=sys.stderr)
        return

    # Call the default exception handler
    sys.__excepthook__(exc_type, exc_value, exc_tb)


# Set the custom exception handler
sys.excepthook = exception_handler


def custom_exc(shell, etype, evalue, tb, tb_offset=None):
    if issubclass(etype, NoPythonStackTrace):
        # Handle BuildError exceptions
        print(evalue, file=sys.stderr)
        return
    shell.showtraceback((etype, evalue, tb), tb_offset=tb_offset)


try:
    ip = get_ipython()  # type: ignore
    # Register your custom exception handler
    ip.set_custom_exc((Exception,), custom_exc)
except NameError:
    # Not in IPython, so we won't set the custom exception handler
    pass



from typing import TYPE_CHECKING

from kirin.exception import NoPythonStackTrace

if TYPE_CHECKING:
    from kirin.ir.nodes.base import IRNode


class HintedError(NoPythonStackTrace):
    def __init__(self, *messages: str, help: str | None = None) -> None:
        super().__init__(*messages)
        self.help = help


class ValidationError(HintedError):
    def __init__(self, node: "IRNode", *messages: str, help: str | None = None) -> None:
        super().__init__(*messages, help=help)
        self.node = node


class TypeCheckError(ValidationError):
    pass


class CompilerError(Exception):
    pass



from kirin.passes import Fold
from kirin.prelude import structural_no_opt
from kirin.rewrite import Walk
from kirin.dialects import py, scf, func


def test_simple_loop_unroll():
    @structural_no_opt
    def simple_loop(x):
        for i in range(3):
            x = x + i
        return x

    fold = Fold(structural_no_opt)
    fold(simple_loop)
    Walk(scf.unroll.ForLoop()).rewrite(simple_loop.code)
    assert len(simple_loop.callable_region.blocks) == 1
    stmts = simple_loop.callable_region.blocks[0].stmts
    assert isinstance(stmts.at(0), py.Constant)
    assert isinstance(stmts.at(1), py.Constant)
    assert isinstance(stmts.at(2), py.Add)
    assert isinstance(stmts.at(3), py.Constant)
    assert isinstance(stmts.at(4), py.Add)
    assert isinstance(stmts.at(5), py.Constant)
    assert isinstance(stmts.at(6), py.Add)
    assert isinstance(stmts.at(7), func.Return)
    assert simple_loop(1) == 4



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, TypeVar, ClassVar, Iterator, Sequence
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.print import Printer, Printable
from kirin.ir.ssa import SSAValue, ResultValue
from kirin.ir.use import Use
from kirin.ir.traits import Trait
from kirin.ir.attrs.abc import Attribute
from kirin.ir.exception import TypeCheckError, ValidationError
from kirin.ir.nodes.base import IRNode
from kirin.ir.nodes.view import MutableSequenceView
from kirin.ir.nodes.block import Block
from kirin.ir.nodes.region import Region

if TYPE_CHECKING:
    from kirin.source import SourceInfo
    from kirin.ir.dialect import Dialect
    from kirin.ir.attrs.types import TypeAttribute
    from kirin.ir.nodes.block import Block
    from kirin.ir.nodes.region import Region


@dataclass
class ArgumentList(
    MutableSequenceView[tuple[SSAValue, ...], "Statement", SSAValue], Printable
):
    """A View object that contains a list of Arguemnts of a Statement.

    Description:
        This is a proxy object that provide safe API to manipulate the arguemnts of a statement.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    def __delitem__(self, idx: int) -> None:
        arg = self.field[idx]
        arg.remove_use(Use(self.node, idx))
        new_args = (*self.field[:idx], *self.field[idx + 1 :])
        self.node._args = new_args
        self.field = new_args

    def set_item(self, idx: int, value: SSAValue) -> None:
        """Set the argument SSAVAlue at the specified index.

        Args:
            idx (int): The index of the item to set.
            value (SSAValue): The value to set.
        """
        args = self.field
        args[idx].remove_use(Use(self.node, idx))
        value.add_use(Use(self.node, idx))
        new_args = (*args[:idx], value, *args[idx + 1 :])
        self.node._args = new_args
        self.field = new_args

    def insert(self, idx: int, value: SSAValue) -> None:
        """Insert the argument SSAValue at the specified index.

        Args:
            idx (int): The index to insert the value.
            value (SSAValue): The value to insert.
        """
        args = self.field
        value.add_use(Use(self.node, idx))
        new_args = (*args[:idx], value, *args[idx:])
        self.node._args = new_args
        self.field = new_args

    def get_slice(self, name: str) -> slice:
        """Get the slice of the arguments.

        Args:
            name (str): The name of the slice.

        Returns:
            slice: The slice of the arguments.
        """
        index = self.node._name_args_slice[name]
        if isinstance(index, int):
            return slice(index, index + 1)
        return index

    def print_impl(self, printer: Printer) -> None:
        printer.print_seq(self.field, delim=", ", prefix="[", suffix="]")


@dataclass
class ResultList(MutableSequenceView[list[ResultValue], "Statement", ResultValue]):
    """A View object that contains a list of ResultValue of a Statement.

    Description:
        This is a proxy object that provide safe API to manipulate the result values of a statement

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    def __setitem__(
        self, idx: int | slice, value: ResultValue | Sequence[ResultValue]
    ) -> None:
        raise NotImplementedError("Cannot set result value directly")

    def __delitem__(self, idx: int) -> None:
        result = self.field[idx]
        del self.field[idx]
        result.delete()

    @property
    def types(self) -> Sequence[TypeAttribute]:
        """Get the result types of the Statement.

        Returns:
            Sequence[TypeAttribute]: type of each result value.
        """
        return [result.type for result in self.field]


@dataclass(repr=False)
class Statement(IRNode["Block"]):
    """The Statment is an instruction in the IR

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    name: ClassVar[str]
    dialect: ClassVar[Dialect | None] = field(default=None, init=False, repr=False)
    traits: ClassVar[frozenset[Trait["Statement"]]] = frozenset()
    _arg_groups: ClassVar[frozenset[str]] = frozenset()

    _args: tuple[SSAValue, ...] = field(init=False)
    _results: list[ResultValue] = field(init=False, default_factory=list)
    successors: list[Block] = field(init=False)
    _regions: list[Region] = field(init=False)
    attributes: dict[str, Attribute] = field(init=False)

    parent: Block | None = field(default=None, init=False, repr=False)
    _next_stmt: Statement | None = field(default=None, init=False, repr=False)
    _prev_stmt: Statement | None = field(default=None, init=False, repr=False)

    # NOTE: This is only for syntax sugar to provide
    # access to args via the properties
    _name_args_slice: dict[str, int | slice] = field(
        init=False, repr=False, default_factory=dict
    )

    @property
    def parent_stmt(self) -> Statement | None:
        """Get the parent statement.

        Returns:
            Statement | None: The parent statement.
        """
        if not self.parent_node:
            return None
        return self.parent_node.parent_stmt

    @property
    def parent_node(self) -> Block | None:
        """Get the parent node.

        Returns:
            Block | None: The parent node.
        """
        return self.parent

    @parent_node.setter
    def parent_node(self, parent: Block | None) -> None:
        """Set the parent Block."""
        from kirin.ir.nodes.block import Block

        self.assert_parent(Block, parent)
        self.parent = parent

    @property
    def parent_region(self) -> Region | None:
        """Get the parent Region.
        Returns:
            Region | None: The parent Region.
        """
        if (p := self.parent_node) is not None:
            return p.parent_node
        return None

    @property
    def parent_block(self) -> Block | None:
        """Get the parent Block.

        Returns:
            Block | None: The parent Block.
        """
        return self.parent_node

    @property
    def next_stmt(self) -> Statement | None:
        """Get the next statement."""
        return self._next_stmt

    @next_stmt.setter
    def next_stmt(self, stmt: Statement) -> None:
        """Set the next statement.

        Note:
            Do not directly call this API. use `stmt.insert_after(self)` instead.

        """
        raise ValueError(
            "Cannot set next_stmt directly, use stmt.insert_after(self) or stmt.insert_before(self)"
        )

    @property
    def prev_stmt(self) -> Statement | None:
        """Get the previous statement."""
        return self._prev_stmt

    @prev_stmt.setter
    def prev_stmt(self, stmt: Statement) -> None:
        """Set the previous statement.

        Note:
            Do not directly call this API. use `stmt.insert_before(self)` instead

        """
        raise ValueError(
            "Cannot set prev_stmt directly, use stmt.insert_after(self) or stmt.insert_before(self)"
        )

    def insert_after(self, stmt: Statement) -> None:
        """Insert the current Statement after the input Statement.

        Args:
            stmt (Statement): Input Statement.

        Example:
            The following example demonstrates how to insert a Statement after another Statement.
            After `insert_after` is called, `stmt1` will be inserted after `stmt2`, which appears in IR in the order (stmt2 -> stmt1)
            ```python
            stmt1 = Statement()
            stmt2 = Statement()
            stmt1.insert_after(stmt2)
            ```
        """
        if self._next_stmt is not None and self._prev_stmt is not None:
            raise ValueError(
                f"Cannot insert before a statement that is already in a block: {self.name}"
            )

        if stmt._next_stmt is not None:
            stmt._next_stmt._prev_stmt = self

        self._prev_stmt = stmt
        self._next_stmt = stmt._next_stmt

        self.parent = stmt.parent
        stmt._next_stmt = self

        if self.parent:
            self.parent._stmt_len += 1

            if self._next_stmt is None:
                self.parent._last_stmt = self

    def insert_before(self, stmt: Statement) -> None:
        """Insert the current Statement before the input Statement.

        Args:
            stmt (Statement): Input Statement.

        Example:
            The following example demonstrates how to insert a Statement before another Statement.
            After `insert_before` is called, `stmt1` will be inserted before `stmt2`, which appears in IR in the order (stmt1 -> stmt2)
            ```python
            stmt1 = Statement()
            stmt2 = Statement()
            stmt1.insert_before(stmt2)
            ```
        """
        if self._next_stmt is not None and self._prev_stmt is not None:
            raise ValueError(
                f"Cannot insert before a statement that is already in a block: {self.name}"
            )

        if stmt._prev_stmt is not None:
            stmt._prev_stmt._next_stmt = self

        self._next_stmt = stmt
        self._prev_stmt = stmt._prev_stmt

        self.parent = stmt.parent
        stmt._prev_stmt = self

        if self.parent:
            self.parent._stmt_len += 1

            if self._prev_stmt is None:
                self.parent._first_stmt = self

    def replace_by(self, stmt: Statement) -> None:
        """Replace the current Statement by the input Statement.

        Args:
            stmt (Statement): Input Statement.
        """
        stmt.insert_before(self)
        for result, old_result in zip(stmt._results, self._results):
            old_result.replace_by(result)
            if old_result.name:
                result.name = old_result.name
        self.delete()

    @property
    def args(self) -> ArgumentList:
        """Get the arguments of the Statement.

        Returns:
            ArgumentList: The arguments View of the Statement.
        """
        return ArgumentList(self, self._args)

    @args.setter
    def args(self, args: Sequence[SSAValue]) -> None:
        """Set the arguments of the Statement.

        Args:
            args (Sequence[SSAValue]): The arguments to set.
        """
        new = tuple(args)
        for idx, arg in enumerate(self._args):
            arg.remove_use(Use(self, idx))
        for idx, arg in enumerate(new):
            arg.add_use(Use(self, idx))
        self._args = new

    @property
    def results(self) -> ResultList:
        """Get the result values of the Statement.

        Returns:
            ResultList: The result values View of the Statement.
        """
        return ResultList(self, self._results)

    @property
    def regions(self) -> list[Region]:
        """Get a list of regions of the Statement.

        Returns:
            list[Region]: The list of regions of the Statement.
        """
        return self._regions

    @regions.setter
    def regions(self, regions: list[Region]) -> None:
        """Set the regions of the Statement."""
        for region in self._regions:
            region._parent = None
        for region in regions:
            region._parent = self
        self._regions = regions

    def drop_all_references(self) -> None:
        """Remove all the dependency that reference/uses this Statement."""
        self.parent = None
        for idx, arg in enumerate(self._args):
            arg.remove_use(Use(self, idx))
        for region in self._regions:
            region.drop_all_references()

    def delete(self, safe: bool = True) -> None:
        """Delete the Statement completely from the IR graph.

        Note:
            This method will detach + remove references of the Statement.

        Args:
            safe (bool, optional): If True, raise error if there is anything that still reference components in the Statement. Defaults to True.
        """
        self.detach()
        self.drop_all_references()
        for result in self._results:
            result.delete(safe=safe)

    def detach(self) -> None:
        """detach the statement from its parent block."""
        if self.parent is None:
            return

        parent: Block = self.parent
        prev_stmt = self.prev_stmt
        next_stmt = self.next_stmt

        if prev_stmt is not None:
            prev_stmt._next_stmt = next_stmt
            self._prev_stmt = None
        else:
            assert (
                parent._first_stmt is self
            ), "Invalid statement, has no prev_stmt but not first_stmt"
            parent._first_stmt = next_stmt

        if next_stmt is not None:
            next_stmt._prev_stmt = prev_stmt
            self._next_stmt = None
        else:
            assert (
                parent._last_stmt is self
            ), "Invalid statement, has no next_stmt but not last_stmt"
            parent._last_stmt = prev_stmt

        self.parent = None
        parent._stmt_len -= 1
        return

    def __post_init__(self):
        assert self.name != ""
        assert isinstance(self.name, str)

    def __init__(
        self,
        *,
        args: Sequence[SSAValue] = (),
        regions: Sequence[Region] = (),
        successors: Sequence[Block] = (),
        attributes: Mapping[str, Attribute] = {},
        results: Sequence[ResultValue] = (),
        result_types: Sequence[TypeAttribute] = (),
        args_slice: Mapping[str, int | slice] = {},
        source: SourceInfo | None = None,
    ) -> None:
        super().__init__()
        """Initialize the Statement.

        Args:
            arsg (Sequence[SSAValue], optional): The arguments of the Statement. Defaults to ().
            regions (Sequence[Region], optional): The regions where the Statement belong to. Defaults to ().
            successors (Sequence[Block], optional): The successors of the Statement. Defaults to ().
            attributes (Mapping[str, Attribute], optional): The attributes of the Statement. Defaults to {}.
            results (Sequence[ResultValue], optional): The result values of the Statement. Defaults to ().
            result_types (Sequence[TypeAttribute], optional): The result types of the Statement. Defaults to ().
            args_slice (Mapping[str, int | slice], optional): The arguments slice of the Statement. Defaults to {}.
            source (SourceInfo | None, optional): The source information of the Statement for debugging/stacktracing. Defaults to None.

        """
        self._args = ()
        self._regions = []
        self._name_args_slice = dict(args_slice)
        self.source = source
        self.args = args

        if results:
            self._results = list(results)
            assert (
                len(result_types) == 0
            ), "expect either results or result_types specified, got both"

        if result_types:
            self._results = [
                ResultValue(self, idx, type=type)
                for idx, type in enumerate(result_types)
            ]

        if not results and not result_types:
            self._results = list(results)

        self.successors = list(successors)
        self.attributes = dict(attributes)
        self.regions = list(regions)

        self.parent = None
        self._next_stmt = None
        self._prev_stmt = None
        self.__post_init__()

    @classmethod
    def from_stmt(
        cls,
        other: Statement,
        args: Sequence[SSAValue] | None = None,
        regions: list[Region] | None = None,
        successors: list[Block] | None = None,
        attributes: dict[str, Attribute] | None = None,
    ) -> Self:
        """Create a similar Statement with new `ResultValue` and without
        attaching to any parent block. This still references to the old successor
        and regions.
        """
        obj = cls.__new__(cls)
        Statement.__init__(
            obj,
            args=args or other._args,
            regions=regions or other._regions,
            successors=successors or other.successors,
            attributes=attributes or other.attributes,
            result_types=[result.type for result in other._results],
            args_slice=other._name_args_slice,
        )
        # inherit the hint:
        for result, other_result in zip(obj._results, other._results):
            result.hints = dict(other_result.hints)

        return obj

    def walk(
        self,
        *,
        reverse: bool = False,
        region_first: bool = False,
        include_self: bool = True,
    ) -> Iterator[Statement]:
        """Traversal the Statements of Regions.

        Args:
            reverse (bool, optional): If walk in the reversed manner. Defaults to False.
            region_first (bool, optional): If the walk should go through the Statement first or the Region of a Statement first. Defaults to False.
            include_self (bool, optional): If the walk should include the Statement itself. Defaults to True.

        Yields:
            Iterator[Statement]: An iterator that yield Statements of Blocks in the Region, in the specified order.
        """
        if include_self and not region_first:
            yield self

        for region in reversed(self.regions) if reverse else self.regions:
            yield from region.walk(reverse=reverse, region_first=region_first)

        if include_self and region_first:
            yield self

    def is_structurally_equal(
        self,
        other: Self,
        context: dict[IRNode | SSAValue, IRNode | SSAValue] | None = None,
    ) -> bool:
        """Check if the Statement is structurally equal to another Statement.

        Args:
            other (Self): The other Statelemt to compare with.
            context (dict[IRNode  |  SSAValue, IRNode  |  SSAValue] | None, optional): A map of IRNode/SSAValue to hint that they are equivalent so the check will treat them as equivalent. Defaults to None.

        Returns:
            bool: True if the IRNode is structurally equal to the other.
        """
        if context is None:
            context = {}

        if self.name != other.name:
            return False

        if (
            len(self.args) != len(other.args)
            or len(self.regions) != len(other.regions)
            or len(self.successors) != len(other.successors)
            or self.attributes != other.attributes
        ):
            return False

        if (
            self.parent is not None
            and other.parent is not None
            and context.get(self.parent) != other.parent
        ):
            return False

        if not all(
            context.get(arg, arg) == other_arg
            for arg, other_arg in zip(self.args, other.args)
        ):
            return False

        if not all(
            context.get(successor, successor) == other_successor
            for successor, other_successor in zip(self.successors, other.successors)
        ):
            return False

        if not all(
            region.is_structurally_equal(other_region, context)
            for region, other_region in zip(self.regions, other.regions)
        ):
            return False

        for result, other_result in zip(self._results, other._results):
            context[result] = other_result

        return True

    def __hash__(self) -> int:
        return id(self)

    def print_impl(self, printer: Printer) -> None:
        from kirin.decl import fields as stmt_fields

        printer.print_name(self)
        printer.plain_print("(")
        for idx, (name, s) in enumerate(self._name_args_slice.items()):
            values = self.args[s]
            if (fields := stmt_fields(self)) and not fields.args[name].print:
                pass
            else:
                with printer.rich(style="orange4"):
                    printer.plain_print(name, "=")

            if isinstance(values, SSAValue):
                printer.print(values)
            else:
                printer.print_seq(values, delim=", ", prefix="(", suffix=")")

            if idx < len(self._name_args_slice) - 1:
                printer.plain_print(", ")

        # NOTE: args are specified manually without names
        if not self._name_args_slice and self._args:
            printer.print_seq(self._args, delim=", ")

        printer.plain_print(")")

        if self.successors:
            printer.print_seq(
                (printer.state.block_id[successor] for successor in self.successors),
                emit=printer.plain_print,
                delim=", ",
                prefix="[",
                suffix="]",
            )

        if self.regions:
            printer.print_seq(
                self.regions,
                delim=" ",
                prefix=" (",
                suffix=")",
            )

        if self.attributes:
            printer.plain_print("{")
            with printer.rich(highlight=True):
                printer.print_mapping(self.attributes, delim=", ")
            printer.plain_print("}")

        if self._results:
            with printer.rich(style="black"):
                printer.plain_print(" : ")
                printer.print_seq(
                    [result.type for result in self._results],
                    delim=", ",
                )

    def get_attr_or_prop(self, key: str) -> Attribute | None:
        """Get the attribute or property of the Statement.

        Args:
            key (str): The key of the attribute or property.

        Returns:
            Attribute | None: The attribute or property of the Statement.
        """
        return self.attributes.get(key)

    @classmethod
    def has_trait(cls, trait_type: type[Trait["Statement"]]) -> bool:
        """Check if the Statement has a specific trait.

        Args:
            trait_type (type[StmtTrait]): The type of trait to check for.

        Returns:
            bool: True if the class has the specified trait, False otherwise.
        """
        for trait in cls.traits:
            if isinstance(trait, trait_type):
                return True
        return False

    TraitType = TypeVar("TraitType", bound=Trait["Statement"])

    @classmethod
    def get_trait(cls, trait: type[TraitType]) -> TraitType | None:
        """Get the trait of the Statement."""
        for t in cls.traits:
            if isinstance(t, trait):
                return t
        return None

    def expect_one_result(self) -> ResultValue:
        """Check if the statement contain only one result, and return it"""
        if len(self._results) != 1:
            raise ValueError(f"expected one result, got {len(self._results)}")
        return self._results[0]

    def check_type(self) -> None:
        """Check the types of the Block. Raises `Exception` if the types are not correct.
        This method is called by the `verify_type` method, which will detect the source
        of the error in the IR. One should always call the `verify_type` method to verify
        the types of the IR.

        Note:
            This method is generated by the `@statement` decorator. But can be overridden
            if needed.
        """
        raise NotImplementedError(
            "check_type should be implemented in the "
            "statement or generated by the @statement decorator"
        )

    def check(self) -> None:
        """Check the statement. Raises `Exception` if the statement is not correct.
        This method is called by the `verify` method, which will detect the source
        of the error in the IR. One should always call the `verify` method to verify
        the IR.

        The difference between `check` and `check_type` is that `check` is called
        at any time to check the structure of the IR by `verify`, while `check_type`
        is called after the type inference to check the types of the IR.
        """
        return

    def verify(self) -> None:
        try:
            self.check()
        except ValidationError as e:
            raise e
        except Exception as e:
            raise ValidationError(self, *e.args) from e

    def verify_type(self) -> None:
        """Verify the type of the statement.

        Note:
            This API should be called after all the types are figured out (by typeinfer)
        """
        try:
            self.check_type()
        except TypeCheckError as e:
            raise e
        except Exception as e:
            raise TypeCheckError(self, *e.args) from e



from typing import Any

from kirin import types
from kirin.prelude import basic
from kirin.dialects.ilist import IList, IListType


@basic
def complicated_type(x: IList[tuple[float, float, IList[float, Any]], Any]):
    return x


def test_complicated_type():
    typ = complicated_type.arg_types[0]
    assert isinstance(typ, types.Generic)
    assert typ.is_subseteq(
        IListType[
            types.Tuple[types.Float, types.Float, IListType[types.Float]], types.Any
        ]
    )



import io

from rich.console import Console

from kirin import ir, types
from kirin.print import Printer
from kirin.prelude import basic
from kirin.dialects import py, func


@basic
def move_gen(start, stop):
    def foo(aod):
        def moo(aod):
            return start, aod

        py.Map(moo, aod)  # type: ignore
        return moo

    return foo(stop)


@basic
def unstable(x: int):  # type: ignore
    y = x + 1
    if y > 10:
        z = y
    else:
        z = y + 1.2
    return z


@basic
def empty():
    pass


class TestBasicPrint:

    def dummy_check(self, node):
        printer = Printer()
        printer.print(node)
        printer.plain_print("\n")

    def check_print(self, node, *text: str):
        printer = Printer()
        with printer.string_io() as stream:
            printer.print(node)
            answer = stream.getvalue()
            for txt in text:
                assert self.rich_str(txt) in answer

    def rich_str(self, text: str):
        try:
            file = io.StringIO()
            console = Console()
            console.file = file
            console.print(text, sep="", end="", highlight=False)
            return file.getvalue()
        finally:
            file.close()

    def test_pytypes(self):
        self.check_print(types.Int, "![dark_blue]py[/dark_blue].int")
        self.check_print(types.Any, "!Any")
        self.check_print(types.Tuple, "![dark_blue]py[/dark_blue].tuple", "~T")
        self.check_print(types.Vararg(types.Int), "*![dark_blue]py[/dark_blue].int")
        self.check_print(
            types.Int,
            "![dark_blue]py[/dark_blue].int",
        )
        self.check_print(
            types.Union(types.Int, types.Float),
            "!Union",
            "![dark_blue]py[/dark_blue].int",
            "![dark_blue]py[/dark_blue].float",
        )

        self.check_print(
            ir.PyAttr(1),
            "1[bright_black] : [/bright_black]",
            "[bright_black]![dark_blue]py[/dark_blue].int[/bright_black]",
        )

        # TODO: actually test these
        self.dummy_check(move_gen)
        self.dummy_check(unstable)
        self.dummy_check(empty)
        self.dummy_check(empty.code)
        assert isinstance(empty.code, func.Function)
        assert isinstance(empty.callable_region, ir.Region)
        region = empty.callable_region
        self.dummy_check(region.blocks[0])
        empty.callable_region.blocks[0].detach()
        self.dummy_check(empty.code)



from kirin import ir
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult

from .stmts import For, Yield, IfElse


class UnusedYield(RewriteRule):
    """Trim unused results from `For` and `IfElse` statements."""

    def scan_unused(self, node: ir.Statement):
        any_unused = False
        uses: list[int] = []
        results: list[ir.ResultValue] = []
        for idx, result in enumerate(node.results):
            if result.uses:
                uses.append(idx)
                results.append(result)
            else:
                any_unused = True
        return any_unused, set(uses), results

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, (For, IfElse)):
            return RewriteResult()

        any_unused, uses, results = self.scan_unused(node)
        if not any_unused:
            return RewriteResult()

        node._results = results
        for region in node.regions:
            for block in region.blocks:
                if not isinstance(block.last_stmt, Yield):
                    continue

                block.last_stmt.args = [block.last_stmt.args[idx] for idx in uses]

        if isinstance(node, For):
            for idx, block_arg in enumerate(node.body.blocks[0].args[1:]):
                if idx not in uses:
                    block_arg.replace_by(node.initializers[idx])
                    block_arg.delete()
            node.initializers = tuple(node.initializers[idx] for idx in uses)
        return RewriteResult(has_done_something=True)



"""The binop dialect for Python.

This module contains the dialect for binary operation semantics in Python, including:

- The `Add`, `Sub`, `Mult`, `Div`, `FloorDiv`, `Mod`, `Pow`,
    `LShift`, `RShift`, `BitOr`, `BitXor`, and `BitAnd` statement classes.
- The lowering pass for binary operations.
- The concrete implementation of binary operations.
- The type inference implementation of binary operations.
- The Julia emitter for binary operations.

This dialect maps `ast.BinOp` nodes to the `Add`, `Sub`, `Mult`, `Div`, `FloorDiv`,
`Mod`, `Pow`, `LShift`, `RShift`, `BitOr`, `BitXor`, and `BitAnd` statements.
"""

from . import (
    julia as julia,
    interp as interp,
    lowering as lowering,
    typeinfer as typeinfer,
)
from .stmts import *  # noqa: F403
from ._dialect import dialect as dialect



import sys
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Generic, TypeVar, ClassVar, Optional, Sequence
from dataclasses import field, dataclass

from typing_extensions import Self, deprecated

from kirin.ir import Block, Region, Statement, DialectGroup, traits
from kirin.ir.method import Method

from .impl import Signature
from .frame import FrameABC
from .state import InterpreterState
from .value import ReturnValue, SpecialValue, StatementResult
from .result import Ok, Err, Result
from .exceptions import InterpreterError

if TYPE_CHECKING:
    from kirin.registry import StatementImpl, InterpreterRegistry

ValueType = TypeVar("ValueType")
FrameType = TypeVar("FrameType", bound=FrameABC)


class InterpreterMeta(ABCMeta):
    """A metaclass for interpreters."""

    pass


@dataclass
class BaseInterpreter(ABC, Generic[FrameType, ValueType], metaclass=InterpreterMeta):
    """A base class for interpreters.

    This class defines the basic structure of an interpreter. It is
    designed to be subclassed to provide the actual implementation of
    the interpreter.

    ### Required Overrides
    When subclassing, if the subclass does not contain `ABC`,
    the subclass must define the following attributes:

    - `keys`: a list of strings that defines the order of dialects to select from.
    - `void`: the value to return when the interpreter evaluates nothing.
    """

    keys: ClassVar[list[str]]
    """The name of the interpreter to select from dialects by order.
    """
    void: ValueType = field(init=False)
    """What to return when the interpreter evaluates nothing.
    """
    dialects: DialectGroup
    """The dialects to interpret.
    """
    fuel: int | None = field(default=None, kw_only=True)
    """The fuel limit for the interpreter.
    """
    debug: bool = field(default=False, kw_only=True)
    """Whether to enable debug mode.
    """
    max_depth: int = field(default=128, kw_only=True)
    """The maximum depth of the interpreter stack.
    """
    max_python_recursion_depth: int = field(default=8192, kw_only=True)
    """The maximum recursion depth of the Python interpreter.
    """

    # global states
    registry: "InterpreterRegistry" = field(init=False, compare=False)
    """The interpreter registry.
    """
    symbol_table: dict[str, Statement] = field(init=False, compare=False)
    """The symbol table.
    """
    state: InterpreterState[FrameType] = field(init=False, compare=False)
    """The interpreter state.
    """

    # private
    _eval_lock: bool = field(default=False, init=False, compare=False)

    def __post_init__(self) -> None:
        self.registry = self.dialects.registry.interpreter(keys=self.keys)

    def initialize(self) -> Self:
        """Initialize the interpreter global states. This method is called right upon
        calling [`run`][kirin.interp.base.BaseInterpreter.run] to initialize the
        interpreter global states.

        !!! note "Default Implementation"
            This method provides default behavior but may be overridden by subclasses
            to customize or extend functionality.
        """
        self.symbol_table: dict[str, Statement] = {}
        self.state: InterpreterState[FrameType] = InterpreterState()
        return self

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if ABC in cls.__bases__:
            return

        if not hasattr(cls, "keys"):
            raise TypeError(f"keys is not defined for class {cls.__name__}")
        if not hasattr(cls, "void"):
            raise TypeError(f"void is not defined for class {cls.__name__}")

    @deprecated("use run instead")
    def eval(
        self,
        mt: Method,
        args: tuple[ValueType, ...],
        kwargs: dict[str, ValueType] | None = None,
    ) -> Result[ValueType]:
        return self.run(mt, args, kwargs)

    def run(
        self,
        mt: Method,
        args: tuple[ValueType, ...],
        kwargs: dict[str, ValueType] | None = None,
    ) -> Result[ValueType]:
        """Run a method. This is the main entry point of the interpreter.

        Args:
            mt (Method): the method to run.
            args (tuple[ValueType]): the arguments to the method, does not include self.
            kwargs (dict[str, ValueType], optional): the keyword arguments to the method.

        Returns:
            Result[ValueType]: the result of the method.
        """
        if self._eval_lock:
            raise InterpreterError(
                "recursive eval is not allowed, use run_method instead"
            )

        self._eval_lock = True
        self.initialize()
        current_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(self.max_python_recursion_depth)
        args = self.get_args(mt.arg_names[len(args) + 1 :], args, kwargs)
        try:
            _, results = self.run_method(mt, args)
        except InterpreterError as e:
            # NOTE: initialize will create new State
            # so we don't need to copy the frames.
            return Err(e, self.state.frames)
        finally:
            self._eval_lock = False
            sys.setrecursionlimit(current_recursion_limit)
        return Ok(results)

    def run_stmt(
        self, stmt: Statement, args: tuple[ValueType, ...]
    ) -> StatementResult[ValueType]:
        """execute a statement with arguments in a new frame.

        Args:
            stmt (Statement): the statement to run.
            args (tuple[ValueType]): the arguments to the statement.

        Returns:
            StatementResult[ValueType]: the result of the statement.
        """
        frame = self.new_frame(stmt)
        self.state.push_frame(frame)
        frame.set_values(stmt.args, args)
        results = self.eval_stmt(frame, stmt)
        self.state.pop_frame()
        return results

    @abstractmethod
    def run_method(
        self, method: Method, args: tuple[ValueType, ...]
    ) -> tuple[FrameType, ValueType]:
        """How to run a method.

        This is defined by subclasses to describe what's the corresponding
        value of a method during the interpretation. Usually, this method
        just calls [`run_callable`][kirin.interp.base.BaseInterpreter.run_callable].

        Args:
            method (Method): the method to run.
            args (tuple[ValueType]): the arguments to the method, does not include self.

        Returns:
            ValueType: the result of the method.
        """
        ...

    def run_callable(
        self, code: Statement, args: tuple[ValueType, ...]
    ) -> tuple[FrameType, ValueType]:
        """Run a callable statement.

        Args:
            code (Statement): the statement to run.
            args (tuple[ValueType]): the arguments to the statement,
                includes self if the corresponding callable region contains a self argument.

        Returns:
            ValueType: the result of the statement.
        """
        if len(self.state.frames) >= self.max_depth:
            return self.eval_recursion_limit(self.state.current_frame())

        interface = code.get_trait(traits.CallableStmtInterface)
        if interface is None:
            raise InterpreterError(f"statement {code.name} is not callable")

        frame = self.new_frame(code)
        self.state.push_frame(frame)
        body = interface.get_callable_region(code)
        if not body.blocks:
            return self.state.pop_frame(), self.void
        frame.set_values(body.blocks[0].args, args)
        results = self.run_callable_region(frame, code, body)
        return self.state.pop_frame(), results

    def run_callable_region(
        self, frame: FrameType, code: Statement, region: Region
    ) -> ValueType:
        """A hook defines how to run the callable region given
        the interpreter context. Frame should be pushed before calling
        this method and popped after calling this method.

        A callable region is a region that can be called as a function.
        Unlike a general region (or the MLIR convention), it always return a value
        to be compatible with the Python convention.
        """
        results = self.run_ssacfg_region(frame, region)
        if isinstance(results, ReturnValue):
            return results.value
        elif not results:  # empty result or None
            return self.void
        raise InterpreterError(
            f"callable region {code.name} does not return `ReturnValue`, got {results}"
        )

    def run_block(self, frame: FrameType, block: Block) -> SpecialValue[ValueType]:
        """Run a block within the current frame.

        Args:
            frame: the current frame.
            block: the block to run.

        Returns:
            SpecialValue: the result of running the block terminator.
        """
        ...

    @abstractmethod
    def new_frame(self, code: Statement) -> FrameType:
        """Create a new frame for the given method."""
        ...

    @staticmethod
    def get_args(
        left_arg_names, args: tuple[ValueType, ...], kwargs: dict[str, ValueType] | None
    ) -> tuple[ValueType, ...]:
        if kwargs:
            # NOTE: #self# is not user input so it is not
            # in the args, +1 is for self
            for name in left_arg_names:
                args += (kwargs[name],)
        return args

    @staticmethod
    def permute_values(
        arg_names: Sequence[str],
        values: tuple[ValueType, ...],
        kwarg_names: tuple[str, ...],
    ) -> tuple[ValueType, ...]:
        """Permute the arguments according to the method signature and
        the given keyword arguments, where the keyword argument names
        refer to the last n arguments in the values tuple.

        Args:
            arg_names: the argument names
            values: the values tuple (should not contain method itself)
            kwarg_names: the keyword argument names
        """
        n_total = len(values)
        if kwarg_names:
            kwargs = dict(zip(kwarg_names, values[n_total - len(kwarg_names) :]))
        else:
            kwargs = None

        positionals = values[: n_total - len(kwarg_names)]
        args = BaseInterpreter.get_args(
            arg_names[len(positionals) + 1 :], positionals, kwargs
        )
        return args

    def eval_stmt(
        self, frame: FrameType, stmt: Statement
    ) -> StatementResult[ValueType]:
        """Run a statement within the current frame. This is the entry
        point of running a statement. It will look up the statement implementation
        in the dialect registry, or optionally call a fallback implementation.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            StatementResult: the result of running the statement

        Note:
            Overload this method for the following reasons:
            - to change the source tracking information
            - to take control of how to run a statement
            - to change the implementation lookup behavior that cannot acheive
                by overloading [`lookup_registry`][kirin.interp.base.BaseInterpreter.lookup_registry]

        Example:
            * implement an interpreter that only handles MyStmt:
            ```python
                class MyInterpreter(BaseInterpreter):
                    ...
                    def eval_stmt(self, frame: FrameType, stmt: Statement) -> StatementResult[ValueType]:
                        if isinstance(stmt, MyStmt):
                            return self.run_my_stmt(frame, stmt)
                        else:
                            return ()
            ```

        """
        # TODO: update tracking information
        method = self.lookup_registry(frame, stmt)
        if method is not None:
            results = method(self, frame, stmt)
            if self.debug and not isinstance(results, (tuple, SpecialValue)):
                raise InterpreterError(
                    f"method must return tuple or SpecialResult, got {results}"
                )
            return results
        elif stmt.dialect not in self.dialects:
            # NOTE: we should terminate the interpreter because this is a
            # deveoper error, not a user error.
            name = stmt.dialect.name if stmt.dialect else "None"
            raise ValueError(f"dialect {name} is not supported by {self.dialects}")

        return self.eval_stmt_fallback(frame, stmt)

    @deprecated("use eval_stmt_fallback instead")
    def run_stmt_fallback(
        self, frame: FrameType, stmt: Statement
    ) -> StatementResult[ValueType]:
        return self.eval_stmt_fallback(frame, stmt)

    def eval_stmt_fallback(
        self, frame: FrameType, stmt: Statement
    ) -> StatementResult[ValueType]:
        """The fallback implementation of statements.

        This is called when no implementation is found for the statement.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            StatementResult: the result of running the statement

        Note:
            Overload this method to provide a fallback implementation for statements.
        """
        # NOTE: not using f-string here because 3.10 and 3.11 have
        #  parser bug that doesn't allow f-string in raise statement
        raise InterpreterError(
            "no implementation for stmt "
            + stmt.print_str(end="")
            + " from "
            + str(type(self))
        )

    def eval_recursion_limit(self, frame: FrameType) -> tuple[FrameType, ValueType]:
        """Return the value of recursion exception, e.g in concrete
        interpreter, it will raise an exception if the limit is reached;
        in type inference, it will return a special value.
        """
        raise InterpreterError("maximum recursion depth exceeded")

    def build_signature(self, frame: FrameType, stmt: Statement) -> "Signature":
        """build signature for querying the statement implementation."""
        return Signature(stmt.__class__, tuple(arg.type for arg in stmt.args))

    def lookup_registry(
        self, frame: FrameType, stmt: Statement
    ) -> Optional["StatementImpl[Self, FrameType]"]:
        """Lookup the statement implementation in the registry.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            Optional[StatementImpl]: the statement implementation if found, None otherwise.
        """
        sig = self.build_signature(frame, stmt)
        if sig in self.registry.statements:
            return self.registry.statements[sig]
        elif (class_sig := Signature(stmt.__class__)) in self.registry.statements:
            return self.registry.statements[class_sig]
        return

    @abstractmethod
    def run_ssacfg_region(
        self, frame: FrameType, region: Region
    ) -> tuple[ValueType, ...] | None | ReturnValue[ValueType]:
        """This implements how to run a region with MLIR SSA CFG convention.

        Args:
            frame: the current frame.
            region: the region to run.

        Returns:
            tuple[ValueType, ...] | SpecialValue[ValueType]: the result of running the region.

        when region returns `tuple[ValueType, ...]`, it means the region terminates normally
        with `YieldValue`. When region returns `ReturnValue`, it means the region terminates
        and needs to pop the frame. Region cannot return `Successor` because reference to
        external region is not allowed.
        """
        ...

    class FuelResult(Enum):
        Stop = 0
        Continue = 1

    def consume_fuel(self) -> FuelResult:
        if self.fuel is None:  # no fuel limit
            return self.FuelResult.Continue

        if self.fuel == 0:
            return self.FuelResult.Stop
        else:
            self.fuel -= 1
            return self.FuelResult.Continue



from .lattice import Value, Bottom, Unknown, PartialTuple, PartialLambda

class _ElemVisitor:
    def is_subseteq_Value(self, other: Value) -> bool: ...
    def is_subseteq_NotConst(self, other: Unknown) -> bool: ...
    def is_subseteq_Unknown(self, other: Bottom) -> bool: ...
    def is_subseteq_PartialTuple(self, other: PartialTuple) -> bool: ...
    def is_subseteq_PartialLambda(self, other: PartialLambda) -> bool: ...
